
"""
datazen - A target-type for grouping other target-tasks together.
"""

# built-in
from typing import List, Tuple

# internal
from datazen.classes.task_environment import TaskEnvironment


class GroupEnvironment(TaskEnvironment):
    """ Leverages a task-environment to group tasks together. """

    def __init__(self):
        """ Add the 'renders' handle. """

        super().__init__()
        self.handles["groups"] = self.valid_group

    def valid_group(self, entry: dict, _: str, dep_data: dict = None,
                    deps_changed: List[str] = None) -> Tuple[bool, bool]:
        """ Stub task to group other tasks. """

        if dep_data is not None:
            self.task_data["groups"][entry["name"]] = dep_data
        return (True, bool(deps_changed))




"""
datazen - A class for adding manifest-loading to environments.
"""

# built-in
import logging
import os
from io import StringIO

# third-party
from cerberus import Validator  # type: ignore

# internal
from datazen.classes.config_environment import ConfigEnvironment
from datazen.classes.template_environment import TemplateEnvironment
from datazen.parsing import load as load_raw
from datazen.parsing import load_stream
from datazen.paths import get_package_data, resolve_dir
from datazen import DEFAULT_DIR, ROOT_NAMESPACE

LOG = logging.getLogger(__name__)


def get_output_dir(data: dict, rel_path: str,
                   default: str = DEFAULT_DIR) -> str:
    """
    Get the resolved output directory based on a dictionary containing
    target data.
    """

    # turn the output directory into a valid path
    out_dir = default
    if "output_dir" in data:
        out_dir = data["output_dir"]
    return resolve_dir(out_dir, rel_path)


def set_output_dir(data: dict, rel_path: str,
                   default: str = DEFAULT_DIR) -> None:
    """ Set the 'output_dir' key correctly on a dictionary. """

    out_dir = get_output_dir(data, rel_path, default)
    os.makedirs(out_dir, exist_ok=True)
    data["output_dir"] = out_dir


class ManifestEnvironment(ConfigEnvironment, TemplateEnvironment):
    """
    A wrapper for the manifest-loading implementations of an environment.
    """

    def __init__(self):
        """ Add a manifest dictionary to the environment. """

        super().__init__()
        self.manifest = {}

    def load_dirs(self, data: dict, rel_path: str,
                  namespace: str = ROOT_NAMESPACE,
                  allow_dup: bool = False) -> None:
        """
        Looks for keys matching types of directories that can be loaded
        into an environment and tries to load them.
        """

        key_handles = {
            "configs": self.add_config_dirs,
            "schemas": self.add_schema_dirs,
            "templates": self.add_template_dirs,
            "variables": self.add_variable_dirs,
        }
        for key in key_handles:
            if key in data:
                key_handles[key](data[key], rel_path, namespace, allow_dup)
            # if a directory list isn't provided, and the directory of the
            # same name of the key is present in the manifest directory,
            # load it
            elif os.path.isdir(os.path.join(rel_path, key)):
                key_handles[key]([key], rel_path, namespace, allow_dup)
            else:
                LOG.debug("not loading any '%s'", key)

    def default_target(self) -> str:
        """
        Retrieve a configured default target for a manifest, otherwise an
        empty String.
        """

        result = ""
        if self.get_valid() and "default_target" in self.manifest["data"]:
            result = self.manifest["data"]["default_target"]
        return result

    def load_manifest(self, path: str = "manifest.yaml") -> bool:
        """ Attempt to load manifest data from a file. """

        # don't allow double-loading manifests
        if self.manifest:
            LOG.error("manifest '%s' already loaded for this environment",
                      self.manifest["path"])
            return False

        self.manifest["path"] = os.path.abspath(path)
        self.manifest["dir"] = os.path.dirname(self.manifest["path"])
        self.manifest["data"], loaded = load_raw(self.manifest["path"], {}, {})

        # make sure we loaded a manifest
        if not loaded:
            self.set_valid(False)
            return self.get_valid()

        # enforce the manifest schema
        schema = get_manifest_schema(False)
        if not schema.validate(self.manifest["data"]):
            LOG.error("invalid manifest: %s", schema.errors)
            self.set_valid(False)
            return self.get_valid()

        # add directories parsed from the schema, paths in the manifest are
        # relative to the directory the manifest is located
        rel_path = self.manifest["dir"]
        set_output_dir(self.manifest["data"], rel_path)

        # load the data directories
        self.load_dirs(self.manifest["data"], rel_path)

        return self.get_valid()


def get_manifest_schema(require_all: bool = True) -> Validator:
    """ Load the schema for manifest from the package. """

    rel_path = os.path.join("schemas", "manifest.yaml")
    schema_str = get_package_data(rel_path)
    return Validator(load_stream(StringIO(schema_str), rel_path)[0],
                     require_all=require_all)




"""
datazen - Tests for the 'schemas' API.
"""

# module under test
from datazen.schemas import validate

# internal
from . import ENV


def test_load_schemas():
    """ Test that the schemas can be loaded. """

    assert ENV.get_schemas(True, True)


def test_validate_configs():
    """ Test that schema data can successfully validate configuration data. """

    schema_data = ENV.get_schemas(True, True)
    assert schema_data
    config_data = ENV.get_configs(True)
    assert config_data

    assert validate(schema_data, config_data)




"""
datazen - Tests for the 'variables' API.
"""

# internal
from . import ENV


def test_load_variables():
    """ Test that the variable data can be loaded. """

    assert ENV.get_variables(True)


def test_load_invalid_variables():
    """ Test that invalid data returns False. """

    assert not ENV.get_variables(False)["bad"]
    assert not ENV.get_configs(False)
    assert not ENV.invalid.configs_valid




"""
datazen - Tests for the 'configs' API.
"""

# internal
from . import ENV


def test_load_configs():
    """ Test that the configuration data can be loaded. """

    config_data = ENV.get_configs(True)

    expected = {"a": "a", "b": "b", "c": "c"}

    assert config_data["a"] == expected
    assert config_data["b"] == expected
    assert config_data["c"] == expected

    assert config_data["d"]["d"] == "d"
    assert config_data["d"]["d"] == "d"
    assert config_data["e"]["e"] == "e"
    assert config_data["f"]["f"] == "f"

    assert config_data["yaml"] == "yaml"
    assert config_data["json"] == "json"




"""
datazen - A dict wrapper that enables simpler schema validation.
"""

# built-in
import logging
from collections import UserDict

# third-party
from cerberus import Validator  # type: ignore

LOG = logging.getLogger(__name__)


class ValidDict(UserDict):
    """
    An object that behaves like a dictionary but can have a provided schema
    enforced.
    """

    def __init__(self, name: str, data: dict, schema: Validator):
        """ Initialize a named, ValidDict. """

        # initialize the dict
        super().__init__(data)

        self.name = name
        self.validator = schema
        self.valid = self.validator.validate(self.data)
        if not self.valid:
            LOG.error("validation error(s) for '%s': %s", self.name,
                      self.validator.errors)
            LOG.error("data: %s", data)




"""
mk - A simpler setuptools-based package definition.
"""

# built-in
import os
from typing import Dict, List

# third-party
import setuptools  # type: ignore


def get_long_description(desc_filename: str = "README.md") -> str:
    """ Get a package's long-description data from a file. """

    with open(desc_filename, "r") as desc_file:
        long_description = desc_file.read()
    return long_description


def get_requirements(reqs_filename: str = None) -> List[str]:
    """ Get a package's requirements based on its requirements file. """

    # use a default file location
    if reqs_filename is None:
        reqs_filename = os.path.join("requirements", "requirements.txt")

    with open(reqs_filename, "r") as reqs_file:
        reqs = reqs_file.read().strip().split()
    return reqs


def get_data_files(pkg_name: str, data_dir: str = "data") -> List[str]:
    """
    Get the non-code sources under a package directory's data directory.
    """

    data_files = []
    for root, _, files in os.walk(os.path.join(pkg_name, data_dir)):
        for fname in files:
            rel_name = os.path.join(root, fname).replace(pkg_name + os.sep, "")
            data_files.append(rel_name)

    return data_files


def setup(pkg_info: Dict[str, str], author_info: Dict[str, str],
          url_override: str = None, entry_override: str = None,
          console_overrides: List[str] = None,
          classifiers_override: List[str] = None) -> None:
    """
    Build a 'setuptools.setup' call with sane defaults and making assumptions
    about certain aspects of a package's structure.
    """

    if entry_override is None:
        entry_override = pkg_info["name"]

    if console_overrides is None:
        entry_str = "{}={}.entry:main".format(entry_override, pkg_info["name"])
        console_overrides = [entry_str]

    if url_override is None:
        url_fstring = "https://github.com/{}/{}"
        url_override = url_fstring.format(author_info["username"],
                                          pkg_info["name"])

    if classifiers_override is None:
        classifiers_override = [
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ]

    setuptools.setup(
        name=pkg_info["name"],
        version=pkg_info["version"],
        author=author_info["name"],
        author_email=author_info["email"],
        description=pkg_info["description"],
        long_description=get_long_description(),
        long_description_content_type="text/markdown",
        url=url_override,
        packages=setuptools.find_packages(),
        classifiers=classifiers_override,
        python_requires=">=3.6",
        entry_points={"console_scripts": console_overrides},
        install_requires=get_requirements(),
        package_data={pkg_info["name"]: get_data_files(pkg_info["name"])},
    )




"""
datazen - Tests for the 'templates' API.
"""

# internal
from . import ENV


def test_load_templates():
    """ Test that the templates can be loaded. """

    template_keys = ["a", "b", "c"]
    templates = ENV.get_templates(True)

    for key in template_keys:
        assert templates[key]
        templates[key].render(ENV.get_configs(True))




"""
datazen - A class for storing metadata about files that have been loaded.
"""

# built-in
from copy import deepcopy
from collections import defaultdict
import logging
import os
import shutil
import time
from typing import Dict, List, Tuple

# internal
from datazen.parsing import set_file_hash, merge
from datazen.load import load_dir_only
from datazen.compile import str_compile

LOG = logging.getLogger(__name__)

DATA_DEFAULT = {"hashes": defaultdict(lambda: defaultdict(dict)),
                "loaded": defaultdict(list)}


class FileInfoCache:
    """ Provides storage for file hashes and lists that have been loaded. """

    def __init__(self, cache_dir: str = None):
        """ Construct an empty cache or optionally load from a directory. """

        self.data: dict = deepcopy(DATA_DEFAULT)
        self.cache_dir: str = ""
        if cache_dir is not None:
            self.load(cache_dir)

    def load(self, cache_dir: str) -> None:
        """ Load data from a directory. """

        assert self.cache_dir == ""
        self.cache_dir = cache_dir
        os.makedirs(self.cache_dir, exist_ok=True)

        # reject things that don't belong by updating instead of assigning
        new_data = load_dir_only(self.cache_dir, True)
        self.data["hashes"].update(new_data["hashes"])
        self.data["loaded"].update(new_data["loaded"])

    def get_hashes(self, sub_dir: str) -> Dict[str, dict]:
        """ Get the cached, dictionary of file hashes for a certain key. """

        return self.data["hashes"][sub_dir]

    def check_hit(self, sub_dir: str, path: str,
                  also_cache: bool = True) -> bool:
        """
        Determine if a given file already exists with its current hash in the
        cache, if not return False and optionally add it to the cache. """

        abs_path = os.path.abspath(path)
        is_new = set_file_hash(self.get_hashes(sub_dir), abs_path, also_cache)
        if also_cache and is_new:
            # guard against a failure in the "new" detection
            loaded_data = self.get_loaded(sub_dir)
            if abs_path not in loaded_data:
                loaded_data.append(abs_path)

        return not is_new

    def get_loaded(self, sub_dir: str) -> List[str]:
        """ Get the cached, list of loaded files for a certain key. """

        return self.data["loaded"][sub_dir]

    def describe(self) -> None:
        """ Describe this cache's contents for debugging purposes. """

        curr_dir = os.getcwd()
        for hash_set, hash_data in self.data["hashes"].items():
            misses = []
            total = 0
            for filename, hash_item in hash_data.items():
                total += 1
                if curr_dir in filename:
                    filename = filename[len(curr_dir) + 1:]
                if not self.check_hit(hash_set, filename, False):
                    misses.append((filename, hash_item))
                else:
                    LOG.debug("%s (%s) matched", filename,
                              time_str(hash_item["time"]))

            # log all misses / updates
            for miss in misses:
                LOG.info("%s (%s) updated", miss[0], time_str(miss[1]["time"]))

            # log a summary
            LOG.info("%s: %d/%d match", hash_set, total - len(misses), total)

    def get_data(self, name: str) -> Tuple[List[str], Dict[str, dict]]:
        """ Get the tuple version of cached data. """

        return (self.get_loaded(name), self.get_hashes(name))

    def clean(self) -> None:
        """ Remove cached data from the file-system. """

        self.data = deepcopy(DATA_DEFAULT)
        if self.cache_dir != "":
            shutil.rmtree(self.cache_dir)
            os.makedirs(self.cache_dir, exist_ok=True)
            LOG.info("cleaning cache at '%s'", self.cache_dir)

    def write(self, out_type: str = "json") -> None:
        """ Commit cached data to the file-system. """

        if self.cache_dir != "":
            for key, val in self.data.items():
                key_path = os.path.join(self.cache_dir,
                                        "{}.{}".format(key, out_type))
                with open(key_path, "w") as key_file:
                    key_file.write(str_compile(val, out_type))
            LOG.debug("wrote cache to '%s'", self.cache_dir)


def copy(cache: FileInfoCache) -> FileInfoCache:
    """ Copy one cache into a new one. """

    new_cache = FileInfoCache()

    # copy the cache
    new_cache.cache_dir = cache.cache_dir
    new_cache.data = deepcopy(cache.data)

    return new_cache


def meld(cache_a: FileInfoCache, cache_b: FileInfoCache) -> None:
    """ Promote all updates from cache_b into cache_a. """

    merge(cache_a.data, cache_b.data, expect_overwrite=True)


def time_str(time_s: float) -> str:
    """ Concert a timestamp to a String. """

    return time.strftime("%c", time.localtime(time_s))


def cmp_loaded_count(cache_a: FileInfoCache, cache_b: FileInfoCache,
                     name: str) -> int:
    """
    Compute the total difference in file counts (for a named group)
    between two caches.
    """

    return abs(len(cache_a.get_loaded(name)) - len(cache_b.get_loaded(name)))


def cmp_total_loaded(cache_a: FileInfoCache, cache_b: FileInfoCache,
                     known_types: List[str]) -> int:
    """
    Compute the total difference in file counts for a provided set of named
    groups.
    """

    result = 0
    for known in known_types:
        result += cmp_loaded_count(cache_a, cache_b, known)
    return result




"""
datazen - APIs for working with file paths.
"""

# built-in
import os
import pkgutil
from typing import List

# internal
from datazen import PKG_NAME


def get_file_name(full_path: str) -> str:
    """ From a full path to a file, get just the name of the file. """

    return ".".join(os.path.basename(full_path).split(".")[:-1])


def get_file_ext(full_path: str) -> str:
    """ From a pull path to a file, get just the file's extension. """

    return os.path.basename(full_path).split(".")[-1]


def get_path_list(root_abs_path: str, current_abs_path: str) -> List[str]:
    """
    From a root directory and a child path, compute the list of directories,
    in order, mapping the root to the child.
    """

    assert len(current_abs_path) >= len(root_abs_path)
    assert root_abs_path in current_abs_path
    return current_abs_path[len(root_abs_path):].split(os.sep)


def advance_dict_by_path(path_list: List[str], data: dict) -> dict:
    """
    Given a dictionary and a list of directory names, return the child
    dictionary advanced by each key, in order, from the provided data.
    """

    for path in path_list:
        if path:
            try:
                data = data[path]
            except KeyError:
                data[path] = {}
                data = data[path]

    return data


def get_package_data(relative_path: str) -> str:
    """ Load a file from this package's data directory. """

    rel_path = os.path.join("data", relative_path)
    schema_raw = pkgutil.get_data(PKG_NAME, rel_path)
    schema_bytes = schema_raw if schema_raw else bytes()
    return schema_bytes.decode("utf-8")


def resolve_dir(data: str, rel_base: str = "") -> str:
    """
    Turn directory data into an absolute path, optionally from a relative
    base.
    """

    data = data.replace("/", os.sep)
    data = data.replace("\\", os.sep)
    if not os.path.isabs(data) and rel_base:
        data = os.path.join(rel_base, data)
    return os.path.abspath(data)




"""
datazen - Useful defaults and other package metadata.
"""

PKG_NAME = "datazen"
VERSION = "1.0.4"
DESCRIPTION = "Compile and render schema-validated configuration data."

DEFAULT_TYPE = "yaml"
DEFAULT_MANIFEST = "manifest.{}".format(DEFAULT_TYPE)
DEFAULT_DIR = "{}-out".format(PKG_NAME)
CACHE_SUFFIX = "_cache"
GLOBAL_KEY = "global"
ROOT_NAMESPACE = "__root__"




"""
datazen - A testing environment class for aggregating commonly used test data.
"""

# built-in
import os
from typing import Dict

# third-party
import jinja2

# modules under test
from datazen.classes.environment import from_manifest

# internal
from .resources import get_resource


class EnvironmentMock:
    """ A class for simple test-data access. """

    def __init__(self):
        """ Initialize data storage. """

        manifest_path = get_resource(os.path.join("manifests", "test.yaml"),
                                     True)
        self.valid = from_manifest(manifest_path)
        assert self.valid.get_valid()

        manifest_path = get_resource(os.path.join("manifests", "valid.yaml"),
                                     False)
        self.invalid = from_manifest(manifest_path)
        assert self.invalid.get_valid()

    def get_configs(self, valid: bool = True) -> dict:
        """ Attempt to load one of the sets of configuration data. """

        env = self.valid if valid else self.invalid
        return env.cached_load_configs()

    def get_schemas(self, valid: bool = True,
                    require_all: bool = True) -> dict:
        """ Attempt to load one of the sets of schemas. """

        env = self.valid if valid else self.invalid
        return env.cached_load_schemas(require_all)

    def get_templates(self, valid: bool = True) -> Dict[str, jinja2.Template]:
        """ Attempt to load one of the sets of templates. """

        env = self.valid if valid else self.invalid
        return env.load_templates()

    def get_variables(self, valid: bool = True) -> dict:
        """ Attempt to load one of the sets of variables. """

        env = self.valid if valid else self.invalid
        return env.cached_load_variables()




"""
datazen - An interface for turning a dictionary into various serialized forms.
"""

# built-in
import json
import logging
import os
from typing import Tuple

# third-party
from ruamel import yaml

# internal
from datazen import DEFAULT_TYPE

LOG = logging.getLogger(__name__)


def str_compile(configs: dict, data_type: str) -> str:
    """
    Serialize dictionary data into the String-form of a specific,
    serializeable type.
    """

    # serialize the data
    if data_type == "json":
        result = json.dumps(configs, indent=4, sort_keys=True)
    elif data_type == "yaml":
        raw_result = yaml.dump(configs)
        result = raw_result if raw_result else ""
    else:
        LOG.error("can't serialize '%s' data (unknown type)", data_type)
        return ""

    return result


def get_compile_output(entry: dict,
                       default_type: str = DEFAULT_TYPE) -> Tuple[str, str]:
    """
    Determine the output path and type of a compile target, from the target's
    data.
    """

    # determine the output type
    output_type = default_type
    if "output_type" in entry:
        output_type = entry["output_type"]

    # write the output
    filename = "{}.{}".format(entry["name"], output_type)
    return os.path.join(entry["output_dir"], filename), output_type




"""
datazen - A command implementation for rendering configuration data (that has
          optionally been resolved from variables) based on provided and
          discovered templates.
"""

# built-in
import logging
from typing import List

# third-party
import jinja2

# internal
from datazen.classes.environment import Environment
from datazen.parsing import load_stream

LOG = logging.getLogger(__name__)


def str_render(template: jinja2.Template, config_data_path: str) -> str:
    """ Load configuration data and render a jinja2 template with it. """

    # load the configuration data from file
    with open(config_data_path) as stream:
        config_data = load_stream(stream, config_data_path)[0]

    return template.render(config_data)


def cmd_render(template_dirs: List[str], template_name: str,
               config_data_path: str, output_file_path: str) -> bool:
    """ Render the desired template from loaded configuration data. """

    env = Environment()

    # add directories
    env.add_template_dirs(template_dirs)

    # load specific template
    template = env.load_templates()[template_name]

    # render the template to the output file, using the new data
    with open(output_file_path, "w") as output:
        output.write(str_render(template, config_data_path))

    LOG.info("'%s' rendered from '%s' (template: '%s')", output_file_path,
             config_data_path, template_name)

    return True




"""
datazen - Top-level APIs for loading and interacting with schema definitions.
"""

# built-in
from typing import List, Dict

# third-party
from cerberus import Validator  # type: ignore

# internal
from datazen.load import load_dir
from datazen.classes.valid_dict import ValidDict


def load(directories: List[str], require_all: bool = True,
         loaded_list: List[str] = None,
         hashes: Dict[str, dict] = None) -> dict:
    """ Load schemas from a list of directories. """

    result: dict = {}

    # load raw data
    for directory in directories:
        load_dir(directory, result, None, loaded_list, hashes)

    # interpret all top-level keys as schemas
    schemas = {}
    for item in result.items():
        schemas[item[0]] = Validator(item[1], require_all=require_all)
    return schemas


def validate(schema_data: Dict[str, Validator], data: dict) -> bool:
    """
    For every top-level key in the schema data, attempt to validate the
    provided data.
    """

    for item in schema_data.items():
        key = item[0]
        if key in data:
            if not ValidDict(key, data[key], item[1]).valid:
                return False

    return True




"""
datazen - A child class for adding schema-data loading capabilities to the
          environment dataset.
"""

# built-in
from typing import List

# internal
from datazen import ROOT_NAMESPACE
from datazen.enums import DataType
from datazen.classes.base_environment import BaseEnvironment, LOADTYPE
from datazen.schemas import load as load_schemas
from datazen.schemas import validate


class SchemaEnvironment(BaseEnvironment):
    """
    The schema-data loading environment mixin, only requires the base
    environment capability to function.
    """

    def load_schemas(self, require_all: bool = True,
                     sch_loads: LOADTYPE = (None, None),
                     name: str = ROOT_NAMESPACE) -> dict:
        """ Load schema data, resolve any un-loaded schema directories. """

        # determine directories that need to be loaded
        data_type = DataType.SCHEMA
        to_load = self.get_to_load(data_type, name)

        # load new data
        schema_data = self.get_data(data_type, name)
        if to_load:
            schema_data.update(load_schemas(to_load, require_all, sch_loads[0],
                                            sch_loads[1]))

        return schema_data

    def enforce_schemas(self, data: dict, require_all: bool = True,
                        sch_loads: LOADTYPE = (None, None),
                        name: str = ROOT_NAMESPACE) -> bool:
        """
        Perform schema-validation on provided data and return the boolean
        result.
        """

        return validate(self.load_schemas(require_all, sch_loads, name), data)

    def add_schema_dirs(self, dir_paths: List[str], rel_path: str = ".",
                        name: str = ROOT_NAMESPACE,
                        allow_dup: bool = False) -> int:
        """
        Add schema-data directories, return the number of directories added.
        """

        return self.add_dirs(DataType.SCHEMA, dir_paths, rel_path, name,
                             allow_dup)




"""
datazen - APIs for loading raw data from files.
"""

# built-in
import hashlib
import io
import json
import logging
from typing import TextIO, List, Tuple
import time

# third-party
import jinja2
from ruamel import yaml  # type: ignore

# internal
from datazen.paths import get_file_ext

LOG = logging.getLogger(__name__)


def get_json_data(data_file: TextIO) -> Tuple[dict, bool]:
    """ Load JSON data from a text stream. """

    data = {}
    loaded = True
    try:
        data = json.load(data_file)
        if not data:
            data = {}
    except json.decoder.JSONDecodeError as exc:
        loaded = False
        LOG.error("json-load error: %s", exc)
    return data, loaded


def get_yaml_data(data_file: TextIO) -> Tuple[dict, bool]:
    """ Load YAML data from a text stream. """

    data = {}
    loaded = True
    try:
        data = yaml.safe_load(data_file)
        if not data:
            data = {}
    except (yaml.scanner.ScannerError, yaml.parser.ParserError) as exc:
        loaded = False
        LOG.error("yaml-load error: %s", exc)
    return data, loaded


def load_stream(data_stream: TextIO, data_path: str) -> Tuple[dict, bool]:
    """
    Load arbitrary data from a text stream, update an existing dictionary.
    """

    # update the dictionary
    ext = get_file_ext(data_path)
    data: dict = {}
    result = False
    if ext == "json":
        data, result = get_json_data(data_stream)
    elif ext == "yaml":
        data, result = get_yaml_data(data_stream)
    else:
        LOG.error("can't load data from '%s' (unknown extension '%s')",
                  data_path, ext)

    if not result:
        LOG.error("failed to load '%s'", data_path)

    return data, result


def merge(dict_a: dict, dict_b: dict, path: List[str] = None,
          expect_overwrite: bool = False) -> dict:
    """
    Combine two dictionaries recursively, prefers dict_a in a conflict. For
    values of the same key that are lists, the lists are combined. Otherwise
    the resulting dictionary is cleanly merged.
    """

    if path is None:
        path = []

    for key in dict_b:
        if key in dict_a:
            # same leaf value
            if dict_a[key] == dict_b[key]:
                pass
            elif isinstance(dict_a[key], dict) and isinstance(dict_b[key],
                                                              dict):
                merge(dict_a[key], dict_b[key], path + [str(key)],
                      expect_overwrite)
            elif isinstance(dict_a[key], list) and isinstance(dict_b[key],
                                                              list):
                dict_a[key].extend(dict_b[key])
            elif not isinstance(dict_a[key], type(dict_b[key])):
                error_str = "Type mismatch at %s" % ".".join(path + [str(key)])
                LOG.error(error_str)
                LOG.error("left:  %s (%s)", type(dict_a[key]), dict_a[key])
                LOG.error("right: %s (%s)", type(dict_b[key]), dict_b[key])
            elif not expect_overwrite:
                error_str = "Conflict at %s" % ".".join(path + [str(key)])
                LOG.error(error_str)
                LOG.error("left:  %s", dict_a[key])
                LOG.error("right: %s", dict_b[key])
            else:
                dict_a[key] = dict_b[key]
        else:
            dict_a[key] = dict_b[key]

    return dict_a


def load(data_path: str, variables: dict,
         dict_to_update: dict,
         expect_overwrite: bool = False) -> Tuple[dict, bool]:
    """
    Load raw file data and meld it into an existing dictionary. Update
    the result as if it's a template using the provided variables.
    """

    # read the raw file and interpret it as a template, resolve 'variables'
    try:
        with open(data_path) as config_file:
            template = jinja2.Template(config_file.read())
            str_output = template.render(variables)
    except FileNotFoundError:
        LOG.error("can't find '%s' to load file data", data_path)
        return ({}, False)

    new_data, loaded = load_stream(io.StringIO(str_output), data_path)
    return merge(dict_to_update, new_data,
                 expect_overwrite=expect_overwrite), loaded


def get_file_hash(path: str) -> str:
    """ Get the MD5 of a file by path. """

    with open(path) as data:
        contents = bytes(data.read(), "utf-8")
    return hashlib.md5(contents).hexdigest()


def set_file_hash(hashes: dict, path: str, set_new: bool = True) -> bool:
    """ Evaluate a hash dictionary and update it on a miss. """

    str_hash = get_file_hash(path)
    result = True
    if path in hashes and str_hash == hashes[path]["hash"]:
        result = False
    elif set_new:
        hashes[path]["hash"] = str_hash
        hashes[path]["time"] = time.time()

    return result




"""
datazen - A child class for adding configuration-data loading capabilities to
          the environment dataset.
"""

# built-in
import logging
from typing import List

# internal
from datazen import ROOT_NAMESPACE
from datazen.enums import DataType
from datazen.classes.base_environment import LOADTYPE
from datazen.classes.variable_environment import VariableEnvironment
from datazen.classes.schema_environment import SchemaEnvironment
from datazen.configs import load as load_configs

LOG = logging.getLogger(__name__)


class ConfigEnvironment(VariableEnvironment, SchemaEnvironment):
    """
    The configuration-data loading environment mixin, requires variable
    loading to function.
    """

    def __init__(self):
        """ Extend the environment with a notion of configs being valid. """

        super().__init__()
        self.configs_valid = False

    def load_configs(self, cfg_loads: LOADTYPE = (None, None),
                     var_loads: LOADTYPE = (None, None),
                     sch_loads: LOADTYPE = (None, None),
                     name: str = ROOT_NAMESPACE) -> dict:
        """
        Load configuration data, resolve any un-loaded configuration
        directories.
        """

        self.configs_valid = False

        # determine directories that need to be loaded
        data_type = DataType.CONFIG
        to_load = self.get_to_load(data_type, name)

        # load new data
        config_data = self.get_data(data_type, name)
        if to_load:
            vdata = self.load_variables(var_loads, name)
            config_data.update(load_configs(to_load, vdata, cfg_loads[0],
                                            cfg_loads[1]))
            self.update_load_state(data_type, to_load, name)

        # enforce schemas
        if not self.enforce_schemas(config_data, True, sch_loads, name):
            LOG.error("schema validation failed, returning an empty dict")
            return {}

        self.configs_valid = True
        return config_data

    def add_config_dirs(self, dir_paths: List[str], rel_path: str = ".",
                        name: str = ROOT_NAMESPACE,
                        allow_dup: bool = False) -> int:
        """
        Add configuration-data directories, return the number of directories
        added.
        """

        return self.add_dirs(DataType.CONFIG, dir_paths, rel_path, name,
                             allow_dup)




"""
datazen - A centralized store for runtime data.
"""

# built-in
from collections import defaultdict
import logging
from typing import Tuple

# internal
from datazen.classes.base_environment import dep_slug_unwrap
from datazen.classes.compile_environment import CompileEnvironment
from datazen.classes.group_environment import GroupEnvironment
from datazen.classes.render_environment import RenderEnvironment

LOG = logging.getLogger(__name__)


class Environment(CompileEnvironment, RenderEnvironment, GroupEnvironment):
    """ A wrapper for inheriting all environment-loading capabilities. """

    def __init__(self):
        """ Add a notion of 'visited' targets to the environment data. """

        super().__init__()
        self.visited = defaultdict(bool)
        self.default = "compiles"

    def execute(self, target: str = "") -> Tuple[bool, bool]:
        """ Execute an arbitrary target. """

        # resolve a default target if one wasn't provided
        data = self.manifest["data"]
        if not target:
            if "default_target" in data:
                target = data["default_target"]
                LOG.info("using default target '%s'", target)
            elif self.default in data and data[self.default]:
                target = data[self.default][0]["name"]
                LOG.info("resolving first target '%s'", target)

        slug = dep_slug_unwrap(target, self.default)
        return self.handle_task(slug[0], slug[1])

    def compile(self, target: str) -> Tuple[bool, bool]:
        """ Execute a named 'compile' target from the manifest. """

        return self.handle_task("compiles", target)

    def render(self, target: str) -> Tuple[bool, bool]:
        """ Execute a named 'render' target from the manifest. """

        return self.handle_task("renders", target)


def from_manifest(manifest_path: str) -> Environment:
    """ Load an environment object from a schema definition on disk. """

    env = Environment()

    # load the manifest
    if not env.load_manifest_with_cache(manifest_path):
        LOG.error("couldn't load manifest at '%s'", manifest_path)

    return env




"""
datazen - A class for adding caching to the manifest-loading environment.
"""

# built-in
import logging
import os
from typing import List, Dict

# third-party
import jinja2

# internal
from datazen.classes.manifest_environment import ManifestEnvironment
from datazen.paths import get_file_name
from datazen import (
    DEFAULT_MANIFEST, CACHE_SUFFIX, ROOT_NAMESPACE
)
from datazen.classes.file_info_cache import FileInfoCache, cmp_total_loaded
from datazen.classes.file_info_cache import copy as copy_cache
from datazen.classes.file_info_cache import meld as meld_cache

LOG = logging.getLogger(__name__)


def manifest_cache_dir(path: str, manifest: dict) -> str:
    """ Find a manifest cache (path) from its path and data. """

    cache_name = ".{}{}".format(get_file_name(path), CACHE_SUFFIX)
    default_cache_dir = os.path.join(manifest["dir"], cache_name)

    # set 'cache_dir' to the default if it wasn't set already
    if "cache_dir" not in manifest["data"]:
        manifest["data"]["cache_dir"] = default_cache_dir

    return os.path.abspath(manifest["data"]["cache_dir"])


class ManifestCacheEnvironment(ManifestEnvironment):
    """ A wrapper for the cache functionality for an environment. """

    def __init__(self):
        """ Extend the environment with a notion of the cache being loaded. """

        super().__init__()
        self.cache = None
        self.aggregate_cache = None
        self.initial_cache = None
        self.manifest_changed = True

    def load_manifest_with_cache(self, path: str = DEFAULT_MANIFEST) -> bool:
        """
        Load a manifest and its cache, or set up a new cache if one doesn't
        exist.
        """
        result = self.load_manifest(path)

        # if we successfully loaded this manifest, try to load its cache
        if result:
            self.cache = FileInfoCache(manifest_cache_dir(path, self.manifest))
            self.aggregate_cache = copy_cache(self.cache)

            # correctly set the state of whether or not this manifest
            # has changed
            self.manifest_changed = not self.cache.check_hit(ROOT_NAMESPACE,
                                                             path)

            # save a copy of the initial cache, so that we can use it to
            # determine if state has changed when evaluating targets
            self.initial_cache = copy_cache(self.cache)
            LOG.debug("cache-environment loaded from '%s'", path)

        return result and self.cache is not None

    def clean_cache(self) -> None:
        """ Remove cached data from the file-system. """

        for name in self.namespaces:
            self.unload_all(name)
        if self.cache is not None:
            self.cache.clean()
        self.manifest_changed = True

    def write_cache(self) -> None:
        """ Commit cached data to the file-system. """

        if self.cache is not None:
            meld_cache(self.aggregate_cache, self.cache)
            self.aggregate_cache.write()

    def describe_cache(self) -> None:
        """ Describe the [initial] cache for debugging purposes. """

        self.initial_cache.describe()

    def restore_cache(self) -> None:
        """ Return the cache to its initially-loaded state. """

        if self.cache is not None:
            meld_cache(self.aggregate_cache, self.cache)
            self.cache = copy_cache(self.initial_cache)

    def get_new_loaded(self, types: List[str]) -> int:
        """
        Compute the number of new files loaded (since the initial load)
        for a set of types;
        """

        return cmp_total_loaded(self.cache, self.initial_cache, types)

    def cached_load_variables(self, name: str = ROOT_NAMESPACE) -> dict:
        """ Load variables, proxied through the cache. """

        return self.load_variables(self.cache.get_data("variables"), name)

    def cached_load_schemas(self, require_all: bool = True,
                            name: str = ROOT_NAMESPACE) -> dict:
        """ Load schemas, proxied through the cache. """

        return self.load_schemas(require_all, self.cache.get_data("schemas"),
                                 name)

    def cached_enforce_schemas(self, data: dict,
                               require_all: bool = True,
                               name: str = ROOT_NAMESPACE) -> bool:
        """ Enforce schemas, proxied through the cache. """

        return self.enforce_schemas(data, require_all,
                                    self.cache.get_data("schemas"), name)

    def cached_load_configs(self, name: str = ROOT_NAMESPACE) -> dict:
        """ Load configs, proxied through the cache. """

        return self.load_configs(self.cache.get_data("configs"),
                                 self.cache.get_data("variables"),
                                 self.cache.get_data("schemas"), name)

    def cached_load_templates(
        self,
        name: str = ROOT_NAMESPACE
    ) -> Dict[str, jinja2.Template]:
        """ Load templates, proxied through the cache. """

        return self.load_templates(self.cache.get_data("templates"), name)




"""
datazen - Package definition for distribution.
"""

# internal
from datazen import PKG_NAME, VERSION, DESCRIPTION
from mk.setup import setup


author_info = {"name": "Vaughn Kottler",
               "email": "vaughnkottler@gmail.com",
               "username": "vkottler"}
pkg_info = {"name": PKG_NAME, "version": VERSION, "description": DESCRIPTION}
setup(pkg_info, author_info, entry_override="dz")




"""
datazen - A base class to be extended for runtime data loading and storing.
"""

# built-in
from collections import defaultdict
import logging
from typing import Dict, List, Tuple
from typing import Optional as Opt

# internal
from datazen import ROOT_NAMESPACE
from datazen.enums import DataType
from datazen.classes.environment_namespace import EnvironmentNamespace, clone

# python3.9 regression: https://github.com/PyCQA/pylint/issues/3882
# pylint: disable=unsubscriptable-object
LOADTYPE = Tuple[Opt[List[str]], Opt[Dict[str, dict]]]
# pylint: enable=unsubscriptable-object

LOG = logging.getLogger(__name__)
SLUG_DELIM = "-"


def get_dep_slug(operation: str, name: str) -> str:
    """ Build a key-slug for an operation's target. """

    return "{}{}{}".format(operation, SLUG_DELIM, name)


def dep_slug_unwrap(slug: str, default_op: str) -> Tuple[str, str]:
    """
    From a slug String, determine the operation + target pair, if the
    operation isn't specified use a default.
    """

    result = (default_op, slug)
    if SLUG_DELIM in slug:
        split = slug.split(SLUG_DELIM, 1)
        result = (split[0], split[1])
    return result


class BaseEnvironment:
    """ The base class for environment loading-and-storing management. """

    def __init__(self, default_ns: str = ROOT_NAMESPACE):
        """
        Manage environments by names, set up a dictionary with a root
        namespace.
        """

        self.namespaces = {}
        self.namespaces[default_ns] = EnvironmentNamespace()

    def add_namespace(self, name: str, clone_root: bool = True) -> None:
        """ Add a new namespace, optionally clone from the existing root. """

        self.namespaces[name] = EnvironmentNamespace()
        if clone_root and name != ROOT_NAMESPACE:
            clone(self.namespaces[ROOT_NAMESPACE], self.namespaces[name])

    def get_valid(self, name: str = ROOT_NAMESPACE) -> bool:
        """ Get the 'valid' flag for a namespace. """

        return self.namespaces[name].valid

    def set_valid(self, value: bool, name: str = ROOT_NAMESPACE) -> None:
        """ Set the 'valid' flag for a namespace. """

        self.namespaces[name].valid = value

    def get_to_load(self, dir_type: DataType,
                    name: str = ROOT_NAMESPACE) -> List[str]:
        """ Proxy for a namespace's get_to_load. """

        return self.namespaces[name].get_to_load(dir_type)

    def get_data(self, dir_type: DataType, name: str = ROOT_NAMESPACE) -> dict:
        """ Get the raw data for a directory type from a namespace. """

        return self.namespaces[name].data[dir_type]

    def unload_all(self, name: str = ROOT_NAMESPACE) -> None:
        """ Unload all of the directories for a namespace. """

        return self.namespaces[name].unload_all()

    def update_load_state(self, dir_type: DataType, to_load: List[str],
                          name: str = ROOT_NAMESPACE) -> int:
        """ Proxy for update_load_state for a namespace. """

        return self.namespaces[name].update_load_state(dir_type, to_load)

    def add_dir(self, dir_type: DataType, dir_path: str,
                rel_path: str = ".", name: str = ROOT_NAMESPACE,
                allow_dup: bool = False) -> bool:
        """ Proxy for add_dir for a namespace. """

        return self.namespaces[name].add_dir(dir_type, dir_path, rel_path,
                                             allow_dup)

    def add_dirs(self, dir_type: DataType, dir_paths: List[str],
                 rel_path: str = ".", name: str = ROOT_NAMESPACE,
                 allow_dup: bool = False) -> int:
        """ Proxy for add_dirs for a namespace. """

        return self.namespaces[name].add_dirs(dir_type, dir_paths, rel_path,
                                              allow_dup)

    def get_namespace(self, key_name: str, target: str,
                      target_data: dict) -> str:
        """
        Determine the namespace that a target should use, in general they
        all should be unique unless they don't load anything new.
        """

        load_deps = defaultdict(list)
        load_deps["compiles"] = ["configs", "schemas", "variables"]
        load_deps["renders"] = ["templates"]

        # add a unique namespace for this target if it loads
        # any new data as to not load any of this data upstream,
        # but still cache it (edge cases here?)
        load_dep_list = load_deps[key_name]
        for load_dep in load_dep_list:
            if load_dep in target_data:
                namespace = get_dep_slug(key_name, target)
                self.add_namespace(namespace)
                return namespace

        # don't make a new namespace if we don't load
        # new data
        return ROOT_NAMESPACE




"""
datazen - An environment extension that exposes rendering capabilities.
"""

# built-in
import logging
import os
from typing import List, Tuple

# third-party
import jinja2

# internal
from datazen.classes.task_environment import TaskEnvironment

LOG = logging.getLogger(__name__)


class RenderEnvironment(TaskEnvironment):
    """ Leverages a cache-equipped environment to render templates. """

    def __init__(self):
        """ Add the 'renders' handle. """

        super().__init__()
        self.handles["renders"] = self.valid_render

    def valid_render(self, entry: dict, namespace: str, dep_data: dict = None,
                     deps_changed: List[str] = None) -> Tuple[bool, bool]:
        """ Perform the render specified by the entry. """

        # determine the output that will be produced
        path = entry["name"]
        if "output_path" in entry:
            path = entry["output_path"]
        path = os.path.join(entry["output_dir"], path)

        # load templates
        templates = self.cached_load_templates(namespace)

        if entry["name"] not in templates:
            LOG.error("no template for key '%s' found, options: %s",
                      entry["name"], list(templates.keys()))
            return False, False

        # if dependencies aren't specified, use config data
        change_criteria = ["templates"]
        if not dep_data and "dependencies" not in entry:
            dep_data = self.cached_load_configs(namespace)
            change_criteria.append("configs")
            LOG.debug("no dependencies loaded for '%s', using config data",
                      entry["name"])

        # determine if we need to perform this render
        if self.already_satisfied(entry["name"], path, change_criteria,
                                  deps_changed):
            LOG.debug("render '%s' satisfied, skipping", entry["name"])
            return True, False

        # render the template
        with open(path, "w") as render_out:
            try:
                render_str = templates[entry["name"]].render(dep_data)
                render_out.write(render_str)

                # save the output into a dict for consistency
                self.task_data["renders"][entry["name"]] = render_str
            except jinja2.exceptions.TemplateError as exc:
                LOG.error("couldn't render '%s' to '%s': %s",
                          entry["name"], path, exc)
                return False, False

        LOG.info("(%s) rendered '%s'", entry["name"], path)

        return True, True




"""
datazen - This package's command-line entry-point.
"""

# built-in
import argparse
import logging
import os
import sys
from typing import List

# internal
from datazen.classes.environment import from_manifest
from datazen import VERSION, DESCRIPTION, DEFAULT_MANIFEST

LOG = logging.getLogger(__name__)


def main(argv: List[str] = None) -> int:
    """ Program entry-point. """

    result = 0

    # fall back on command-line arguments
    command_args = sys.argv
    if argv is not None:
        command_args = argv

    # initialize argument parsing
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument("--version", action="version",
                        version="%(prog)s {0}".format(VERSION))
    parser.add_argument("-v", "--verbose", action="store_true",
                        help="set to increase logging verbosity")
    parser.add_argument("-m", "--manifest", default=DEFAULT_MANIFEST,
                        help=("manifest to execute tasks from (default: " +
                              "'%(default)s')"))
    parser.add_argument("-c", "--clean", action="store_true",
                        help="clean the manifest's cache and exit")
    parser.add_argument("-d", "--describe", action="store_true",
                        help="describe the manifest's cache and exit")
    parser.add_argument("-C", "--dir", default=os.getcwd(), dest="dir",
                        help=("execute from a specific directory (default: " +
                              "'%(default)s')"))
    parser.add_argument("targets", nargs="*", help="target(s) to execute")

    starting_dir = os.getcwd()

    # parse arguments and execute the requested command
    try:
        args = parser.parse_args(command_args[1:])
        args.version = VERSION

        # initialize logging
        log_level = logging.DEBUG if args.verbose else logging.INFO
        logging.basicConfig(level=log_level,
                            format=("%(name)-36s - %(levelname)-6s - "
                                    "%(message)s"))

        # change to the specified directory
        os.chdir(args.dir)

        env = from_manifest(args.manifest)

        if env.get_valid():
            # clean, if requested
            if args.clean:
                env.clean_cache()
            elif args.describe:
                env.describe_cache()
            else:
                # execute targets
                if not args.targets:
                    args.targets.append(env.default_target())
                for target in args.targets:
                    task_result = env.execute(target)
                    if not task_result[0]:
                        LOG.error("target '%s' failed", target)
                        result = 1
                        break
                    if not task_result[1]:
                        LOG.info("'%s' already satisfied", target)
        else:
            result = 1

    except SystemExit as exc:
        result = 1
        if exc.code is not None:
            result = exc.code

    # return to starting dir
    os.chdir(starting_dir)

    return result




"""
datazen - Test functions in the 'parsing' module.
"""

# module under test
from datazen.parsing import merge


def test_bad_overwrite():
    """ Test that if we don't want to overwrite, we don't. """

    dict_a = {"a": "a"}
    dict_b = {"a": "b"}
    assert merge(dict_a, dict_b) == dict_a
